import anyio
import asyncclick as click
from bleak import (AdvertisementData, BleakClient, BleakGATTCharacteristic,
                   BleakScanner, BLEDevice)
from bleak.uuids import uuidstr_to_str

from .const import FotaService, ScanService
from .parse import Characteristic


@click.group()
async def main():
    pass


@main.command()
async def scan():
    click.echo("Scanning for devices")

    devices = set()

    def detected(device: BLEDevice, advertisement: AdvertisementData):
        if device not in devices:
            if ScanService not in advertisement.service_uuids:
                return
            devices.add(device)

        click.echo(f"Device: {device}")
        for service in advertisement.service_uuids:
            click.echo(f" - Service: {service} {uuidstr_to_str(service)}")
        click.echo(f" - Data: {advertisement.service_data}")
        click.echo(f" - Manu: {advertisement.manufacturer_data}")
        click.echo(f" - RSSI: {advertisement.rssi}")
        click.echo()

    async with BleakScanner(detected, service_uuids=[ScanService, FotaService]):
        await anyio.sleep_forever()


@main.command()
@click.argument("address")
async def connect(address: str):
    click.echo(f"Connecting to: {address}")
    async with BleakClient(address, timeout=20) as client:
        for service in client.services:
            click.echo(f"Service: {service}")

            async def read_print(char: BleakGATTCharacteristic):
                parser = Characteristic.registry.get(char.uuid)
                if "read" in char.properties:
                    data = await client.read_gatt_char(char.uuid)
                else:
                    data = None
                click.echo(f" -  {char}")
                click.echo(f" -  {char.properties}")
                if data is not None and parser:
                    click.echo(f" -  Data: {parser.decode(data)}")

            async with anyio.create_task_group() as tg:
                for char in service.characteristics:
                    tg.start_soon(read_print, char)


@main.command()
async def chars():
    for char in Characteristic.registry.values():
        click.echo(char.name)


try:
    main()
except KeyboardInterrupt:
    pass



from __future__ import annotations

import logging
from collections.abc import Iterable
from datetime import datetime

from bleak import BleakClient
from bleak.uuids import register_uuids

from .const import DeviceConfiguration, ScanService
from .exceptions import CharacteristicNoAccess, CharacteristicNotFound
from .parse import Characteristic, CharacteristicType, Service

LOGGER = logging.getLogger(__name__)

register_uuids(
    {
        service.uuid: f"Gardena {service.__name__}"
        for service in Service.registry.values()
    }
)

register_uuids(
    {
        char.uuid: f"Gardena {service.__name__} {char.name}"
        for service in Service.registry.values()
        for char in service.characteristics()
    }
)

register_uuids({ScanService: "Husqvarna"})


async def read_char_raw(client: BleakClient, uuid: str):
    characteristic = client.services.get_characteristic(uuid)
    if characteristic is None:
        raise CharacteristicNotFound(f"Unable to find characteristic {uuid}")
    if "read" not in characteristic.properties:
        raise CharacteristicNoAccess(f"Characteristic {uuid} is not readable")
    return await client.read_gatt_char(characteristic)


async def read_char(
    client: BleakClient, char: Characteristic[CharacteristicType]
) -> CharacteristicType:
    """Read data to from a characteristic."""
    return char.decode(await read_char_raw(client, char.uuid))


async def write_char_raw(
    client: BleakClient, uuid: str, data: bytes, response: bool = True
):
    """Write data to a characteristic."""
    characteristic = client.services.get_characteristic(uuid)
    if characteristic is None:
        raise CharacteristicNotFound(f"Unable to find characteristic {uuid}")
    if "write" not in characteristic.properties:
        raise CharacteristicNoAccess(f"Characteristic {uuid} is not writable")
    await client.write_gatt_char(characteristic, data, response=response)


async def write_char(
    client: BleakClient,
    char: Characteristic[CharacteristicType],
    value: CharacteristicType,
    response=True,
) -> None:
    """Write data to a characteristic."""
    data = char.encode(value)
    write_char_raw(client, char.uuid, data, response)


async def update_timestamp(client: BleakClient, now: datetime):
    timestamp = await read_char(client, DeviceConfiguration.unix_timestamp)
    timestamp = timestamp.replace(tzinfo=now.tzinfo)
    delta = timestamp - now
    if abs(delta.total_seconds()) > 60:
        LOGGER.warning(
            "Updating time on device to match local time delta was %s", delta
        )
        await write_char(
            client,
            DeviceConfiguration.unix_timestamp,
            now.replace(tzinfo=None),
            True,
        )


async def get_all_characteristics_uuid(client: BleakClient) -> set[str]:
    """Get all characteristics from device."""
    return {
        characteristic.uuid
        for service in client.services
        for characteristic in service.characteristics
    }
